.dropdown-toggle {
    font-size: 1rem !important;
    padding: .5rem !important;
}

.header-dropdown-col {
    padding: 0rem 0rem .25rem 0rem;
}

.header-left-col {
    padding: 0rem .25rem 0rem 0rem;
}

.header-right-col {
    padding: 0rem 0rem 0rem .5rem;
}

.header-row {
    margin: 0rem;
    padding: 1rem 0rem 0rem .25rem;
    position: relative;
    width: 100%;
    z-index: 100;
}

.header-text {
    color: rgb(40, 60, 70);
    padding-top: .7rem;
}
